@@ -283,6 +283,7 @@ parse_filter(int c, gnc_t gnc, void *closure, struct filter **filter_return)
     if(filter == NULL)
         goto error;
     filter->plen_le = 128;
+    filter->src_plen_le = 128;
 
     while(1) {
         c = skip_whitespace(c, gnc, closure);
@@ -295,10 +296,25 @@ parse_filter(int c, gnc_t gnc, void *closure, struct filter **filter_return)
             goto error;
 
         if(strcmp(token, "ip") == 0) {
-            c = getnet(c, &filter->prefix, &filter->plen, &filter->af,
+            int af;
+            c = getnet(c, &filter->prefix, &filter->plen, &af,
                        gnc, closure);
             if(c < -1)
                 goto error;
+            if(filter->af == AF_UNSPEC)
+                filter->af = af;
+            else if(filter->af != af)
+                goto error;
+        } else if(strcmp(token, "src-ip") == 0) {
+            int af;
+            c = getnet(c, &filter->src_prefix, &filter->src_plen, &af,
+                       gnc, closure);
+            if(c < -1)
+                goto error;
+            if(filter->af == AF_UNSPEC)
+                filter->af = af;
+            else if(filter->af != af)
+                goto error;
         } else if(strcmp(token, "eq") == 0) {
             int p;
             c = getint(c, &p, gnc, closure);
@@ -318,6 +334,25 @@ parse_filter(int c, gnc_t gnc, void *closure, struct filter **filter_return)
             if(c < -1)
                 goto error;
             filter->plen_ge = MAX(filter->plen_ge, p);
+        } else if(strcmp(token, "src-eq") == 0) {
+            int p;
+            c = getint(c, &p, gnc, closure);
+            if(c < -1)
+                goto error;
+            filter->src_plen_ge = MAX(filter->src_plen_ge, p);
+            filter->src_plen_le = MIN(filter->src_plen_le, p);
+        } else if(strcmp(token, "src-le") == 0) {
+            int p;
+            c = getint(c, &p, gnc, closure);
+            if(c < -1)
+                goto error;
+            filter->src_plen_le = MIN(filter->src_plen_le, p);
+        } else if(strcmp(token, "src-ge") == 0) {
+            int p;
+            c = getint(c, &p, gnc, closure);
+            if(c < -1)
+                goto error;
+            filter->src_plen_ge = MAX(filter->src_plen_ge, p);
         } else if(strcmp(token, "neigh") == 0) {
             unsigned char *neigh = NULL;
             c = getip(c, &neigh, NULL, gnc, closure);
@@ -346,23 +381,36 @@ parse_filter(int c, gnc_t gnc, void *closure, struct filter **filter_return)
             filter->ifname = interface;
             filter->ifindex = if_nametoindex(interface);
         } else if(strcmp(token, "allow") == 0) {
-            filter->result = 0;
+            filter->action.add_metric = 0;
         } else if(strcmp(token, "deny") == 0) {
-            filter->result = INFINITY;
+            filter->action.add_metric = INFINITY;
         } else if(strcmp(token, "metric") == 0) {
             int metric;
             c = getint(c, &metric, gnc, closure);
             if(c < -1) goto error;
             if(metric <= 0 || metric > INFINITY)
                 goto error;
-            filter->result = metric;
+            filter->action.add_metric = metric;
+        } else if(strcmp(token, "src-prefix") == 0) {
+            int af;
+            c = getnet(c, &filter->action.src_prefix, &filter->action.src_plen,
+                       &af, gnc, closure);
+            if(c < -1)
+                goto error;
+            if(filter->af == AF_UNSPEC)
+                filter->af = af;
+            else if(filter->af != af)
+                goto error;
+            if(af == AF_INET && filter->action.src_plen == 96)
+                memset(&filter->action.src_prefix, 0, 16);
         } else {
             goto error;
         }
         free(token);
     }
     if(filter->af == 0) {
-        if(filter->plen_le < 128 || filter->plen_ge > 0)
+        if(filter->plen_le < 128 || filter->plen_ge > 0 ||
+           filter->src_plen_le < 128 || filter->src_plen_ge > 0)
             filter->af = AF_INET6;
     } else if(filter->af == AF_INET) {
         filter->plen_le += 96;
@@ -642,7 +690,9 @@ parse_option(int c, gnc_t gnc, void *closure, char *token)
     } else if(strcmp(token, "keep-unfeasible") == 0 ||
               strcmp(token, "link-detect") == 0 ||
               strcmp(token, "random-id") == 0 ||
-              strcmp(token, "daemonise") == 0) {
+              strcmp(token, "daemonise") == 0 ||
+              strcmp(token, "ipv6-subtrees") == 0 ||
+              strcmp(token, "reflect-kernel-metric") == 0) {
         int b;
         c = getbool(c, &b, gnc, closure);
         if(c < -1)
@@ -656,6 +706,10 @@ parse_option(int c, gnc_t gnc, void *closure, char *token)
             random_id = b;
         else if(strcmp(token, "daemonise") == 0)
             do_daemonise = b;
+        else if(strcmp(token, "ipv6-subtrees") == 0)
+            has_ipv6_subtrees = b;
+        else if(strcmp(token, "reflect-kernel-metric") == 0)
+            reflect_kernel_metric = b;
         else
             abort();
     } else if(strcmp(token, "protocol-group") == 0) {
@@ -713,6 +767,18 @@ parse_option(int c, gnc_t gnc, void *closure, char *token)
         if(c < -1 || h < 0)
             goto error;
         change_smoothing_half_life(h);
+    } else if(strcmp(token, "first-table-number") == 0) {
+        int n;
+        c = getint(c, &n, gnc, closure);
+        if(c < -1 || n <= 0 || n + SRC_TABLE_NUM >= 254)
+            goto error;
+        src_table_idx = n;
+    } else if(strcmp(token, "first-rule-priority") == 0) {
+        int n;
+        c = getint(c, &n, gnc, closure);
+        if(c < -1 || n <= 0 || n + SRC_TABLE_NUM >= 32765)
+            goto error;
+        src_table_prio = n;
     } else {
         goto error;
     }
@@ -873,6 +939,7 @@ renumber_filters()
 static int
 filter_match(struct filter *f, const unsigned char *id,
              const unsigned char *prefix, unsigned short plen,
+             const unsigned char *src_prefix, unsigned short src_plen,
              const unsigned char *neigh, unsigned int ifindex, int proto)
 {
     if(f->af) {
@@ -890,6 +957,11 @@ filter_match(struct filter *f, const unsigned char *id,
         if(!prefix || plen < f->plen || !in_prefix(prefix, f->prefix, f->plen))
             return 0;
     }
+    if(f->src_prefix) {
+        if(!src_prefix || src_plen < f->src_plen ||
+           !in_prefix(src_prefix, f->src_prefix, f->src_plen))
+            return 0;
+    }
     if(f->plen_ge > 0 || f->plen_le < 128) {
         if(!prefix)
             return 0;
@@ -898,6 +970,14 @@ filter_match(struct filter *f, const unsigned char *id,
         if(plen < f->plen_ge)
             return 0;
     }
+    if(f->src_plen_ge > 0 || f->src_plen_le < 128) {
+        if(!src_prefix)
+            return 0;
+        if(src_plen > f->src_plen_le)
+            return 0;
+        if(src_plen < f->src_plen_ge)
+            return 0;
+    }
     if(f->neigh) {
         if(!neigh || memcmp(f->neigh, neigh, 16) != 0)
             return 0;
@@ -925,34 +1005,49 @@ filter_match(struct filter *f, const unsigned char *id,
 static int
 do_filter(struct filter *f, const unsigned char *id,
           const unsigned char *prefix, unsigned short plen,
-          const unsigned char *neigh, unsigned int ifindex, int proto)
+          const unsigned char *src_prefix, unsigned short src_plen,
+          const unsigned char *neigh, unsigned int ifindex, int proto,
+          struct filter_result *result)
 {
+    if(result)
+        memset(result, 0, sizeof(struct filter_result));
+
     while(f) {
-        if(filter_match(f, id, prefix, plen, neigh, ifindex, proto))
-            return f->result;
+        if(filter_match(f, id, prefix, plen, src_prefix, src_plen,
+                        neigh, ifindex, proto)) {
+            if(result)
+                memcpy(result, &f->action, sizeof(struct filter_result));
+            return f->action.add_metric;
+        }
         f = f->next;
     }
+
     return -1;
 }
 
 int
 input_filter(const unsigned char *id,
              const unsigned char *prefix, unsigned short plen,
+             const unsigned char *src_prefix, unsigned short src_plen,
              const unsigned char *neigh, unsigned int ifindex)
 {
     int res;
-    res = do_filter(input_filters, id, prefix, plen, neigh, ifindex, 0);
+    res = do_filter(input_filters, id, prefix, plen,
+                    src_prefix, src_plen, neigh, ifindex, 0, NULL);
     if(res < 0)
         res = 0;
     return res;
 }
 
 int
-output_filter(const unsigned char *id, const unsigned char *prefix,
-              unsigned short plen, unsigned int ifindex)
+output_filter(const unsigned char *id,
+              const unsigned char *prefix, unsigned short plen,
+              const unsigned char *src_prefix, unsigned short src_plen,
+              unsigned int ifindex)
 {
     int res;
-    res = do_filter(output_filters, id, prefix, plen, NULL, ifindex, 0);
+    res = do_filter(output_filters, id, prefix, plen,
+                    src_prefix, src_plen, NULL, ifindex, 0, NULL);
     if(res < 0)
         res = 0;
     return res;
@@ -960,11 +1055,13 @@ output_filter(const unsigned char *id, const unsigned char *prefix,
 
 int
 redistribute_filter(const unsigned char *prefix, unsigned short plen,
-                    unsigned int ifindex, int proto)
+                    const unsigned char *src_prefix, unsigned short src_plen,
+                    unsigned int ifindex, int proto,
+                    struct filter_result *result)
 {
     int res;
-    res = do_filter(redistribute_filters, NULL, prefix, plen, NULL,
-                    ifindex, proto);
+    res = do_filter(redistribute_filters, NULL, prefix, plen,
+                    src_prefix, src_plen, NULL, ifindex, proto, result);
     if(res < 0)
         res = INFINITY;
     return res;
@@ -979,6 +1076,7 @@ finalise_config()
 
     filter->proto = RTPROT_BABEL_LOCAL;
     filter->plen_le = 128;
+    filter->src_plen_le = 128;
     add_filter(filter, &redistribute_filters);
 
     while(interface_confs) {
