@@ -40,6 +40,7 @@ THE SOFTWARE.
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
 #include <linux/if_bridge.h>
+#include <linux/fib_rules.h>
 #include <netinet/ether.h>
 
 #if(__GLIBC__ < 2) || (__GLIBC__ == 2 && __GLIBC_MINOR__ <= 5)
@@ -77,6 +78,13 @@ static int dgram_socket = -1;
 #define NO_ARPHRD
 #endif
 
+
+static int find_table(const unsigned char *src, unsigned short src_plen);
+static void release_tables(void);
+static int filter_kernel_rules(struct nlmsghdr *nh, void *data);
+static void install_missing_rules(char rule_exists[SRC_TABLE_NUM], int v4);
+
+
 /* Determine an interface's hardware address, in modified EUI-64 format */
 int
 if_eui64(char *ifname, int ifindex, unsigned char *eui)
@@ -548,6 +556,7 @@ kernel_setup(int setup)
 
         return 1;
     } else {
+        release_tables();
         close(dgram_socket);
         dgram_socket = -1;
 
@@ -613,9 +622,13 @@ kernel_setup_socket(int setup)
                           | rtnlgrp_to_mask(RTNLGRP_IPV4_ROUTE)
                           | rtnlgrp_to_mask(RTNLGRP_LINK)
                           | rtnlgrp_to_mask(RTNLGRP_IPV4_IFADDR)
-                          | rtnlgrp_to_mask(RTNLGRP_IPV6_IFADDR));
+                          | rtnlgrp_to_mask(RTNLGRP_IPV6_IFADDR)
+    /* We monitor rules, because it can be change by third parties.  For example
+       a /etc/init.d/network restart on OpenWRT flush the rules. */
+                          | rtnlgrp_to_mask(RTNLGRP_IPV4_RULE)
+                          | rtnlgrp_to_mask(RTNLGRP_IPV6_RULE));
         if(rc < 0) {
-            perror("netlink_socket(_ROUTE | _LINK | _IFADDR)");
+            perror("netlink_socket(_ROUTE | _LINK | _IFADDR | _RULE)");
             kernel_socket = -1;
             return -1;
         }
@@ -906,16 +919,16 @@ kernel_interface_channel(const char *ifname, int ifindex)
 
 int
 kernel_route(int operation, const unsigned char *dest, unsigned short plen,
+             const unsigned char *src, unsigned short src_plen,
              const unsigned char *gate, int ifindex, unsigned int metric,
              const unsigned char *newgate, int newifindex,
              unsigned int newmetric)
 {
-
     union { char raw[1024]; struct nlmsghdr nh; } buf;
     struct rtmsg *rtm;
     struct rtattr *rta;
     int len = sizeof(buf.raw);
-    int rc, ipv4;
+    int rc, ipv4, table, use_src = 0;
 
     if(!nl_setup) {
         fprintf(stderr,"kernel_route: netlink not initialized.\n");
@@ -937,19 +950,18 @@ kernel_route(int operation, const unsigned char *dest, unsigned short plen,
 
     /* Check that the protocol family is consistent. */
     if(plen >= 96 && v4mapped(dest)) {
-        if(!v4mapped(gate)) {
+        if(!v4mapped(gate) ||
+           (src_plen > 0 && (!v4mapped(src) || src_plen < 96))) {
             errno = EINVAL;
             return -1;
         }
     } else {
-        if(v4mapped(gate)) {
+        if(v4mapped(gate)|| (src_plen > 0 && v4mapped(src))) {
             errno = EINVAL;
             return -1;
         }
     }
 
-    ipv4 = v4mapped(gate);
-
     if(operation == ROUTE_MODIFY) {
         if(newmetric == metric && memcmp(newgate, gate, 16) == 0 &&
            newifindex == ifindex)
@@ -961,9 +973,11 @@ kernel_route(int operation, const unsigned char *dest, unsigned short plen,
            stick with the naive approach, and hope that the window is
            small enough to be negligible. */
         kernel_route(ROUTE_FLUSH, dest, plen,
+                     src, src_plen,
                      gate, ifindex, metric,
                      NULL, 0, 0);
         rc = kernel_route(ROUTE_ADD, dest, plen,
+                          src, src_plen,
                           newgate, newifindex, newmetric,
                           NULL, 0, 0);
         if(rc < 0) {
@@ -975,11 +989,26 @@ kernel_route(int operation, const unsigned char *dest, unsigned short plen,
         return rc;
     }
 
-    kdebugf("kernel_route: %s %s/%d metric %d dev %d nexthop %s\n",
-           operation == ROUTE_ADD ? "add" :
-           operation == ROUTE_FLUSH ? "flush" : "???",
-           format_address(dest), plen, metric, ifindex,
-           format_address(gate));
+
+    ipv4 = v4mapped(gate);
+
+    if(src_plen == 0) {
+        table = export_table;
+    } else if(has_ipv6_subtrees && !ipv4) {
+        table = export_table;
+        use_src = 1;
+    } else {
+        table = find_table(src, src_plen);
+        if(table < 0)
+            return -1;
+    }
+
+    kdebugf("kernel_route: %s %s from %s "
+            "table %d metric %d dev %d nexthop %s\n",
+            operation == ROUTE_ADD ? "add" :
+            operation == ROUTE_FLUSH ? "flush" : "???",
+            format_prefix(dest, plen), format_prefix(src, src_plen),
+            table, metric, ifindex, format_address(gate));
 
     /* Unreachable default routes cause all sort of weird interactions;
        ignore them. */
@@ -998,7 +1027,9 @@ kernel_route(int operation, const unsigned char *dest, unsigned short plen,
     rtm = NLMSG_DATA(&buf.nh);
     rtm->rtm_family = ipv4 ? AF_INET : AF_INET6;
     rtm->rtm_dst_len = ipv4 ? plen - 96 : plen;
-    rtm->rtm_table = export_table;
+    if(use_src)
+        rtm->rtm_src_len = src_plen;
+    rtm->rtm_table = table;
     rtm->rtm_scope = RT_SCOPE_UNIVERSE;
     if(metric < KERNEL_INFINITY)
         rtm->rtm_type = RTN_UNICAST;
@@ -1019,6 +1050,12 @@ kernel_route(int operation, const unsigned char *dest, unsigned short plen,
         rta->rta_len = RTA_LENGTH(sizeof(struct in6_addr));
         rta->rta_type = RTA_DST;
         memcpy(RTA_DATA(rta), dest, sizeof(struct in6_addr));
+        if(use_src) {
+            rta = RTA_NEXT(rta, len);
+            rta->rta_len = RTA_LENGTH(sizeof(struct in6_addr));
+            rta->rta_type = RTA_SRC;
+            memcpy(RTA_DATA(rta), src, sizeof(struct in6_addr));
+        }
     }
 
     rta = RTA_NEXT(rta, len);
@@ -1058,19 +1095,16 @@ parse_kernel_route_rta(struct rtmsg *rtm, int len, struct kernel_route *route)
     struct rtattr *rta= RTM_RTA(rtm);;
     len -= NLMSG_ALIGN(sizeof(*rtm));
 
-    memset(&route->prefix, 0, sizeof(struct in6_addr));
-    memset(&route->gw, 0, sizeof(struct in6_addr));
-    route->plen = rtm->rtm_dst_len;
+    memset(route, 0, sizeof(struct kernel_route));
     if(rtm->rtm_family == AF_INET) {
+        /* if RTA_DST is not a TLV, that's a default destination */
         const unsigned char zeroes[4] = {0, 0, 0, 0};
         v4tov6(route->prefix, zeroes);
-        route->plen += 96;
+        route->plen = 96;
     }
-
-    route->metric = 0;
-    route->ifindex = 0;
     route->proto = rtm->rtm_protocol;
 
+#define GET_PLEN(p) (rtm->rtm_family == AF_INET) ? p + 96 : p
 #define COPY_ADDR(d, s) \
     do { \
         if(rtm->rtm_family == AF_INET6) \
@@ -1084,8 +1118,13 @@ parse_kernel_route_rta(struct rtmsg *rtm, int len, struct kernel_route *route)
     while(RTA_OK(rta, len)) {
         switch(rta->rta_type) {
         case RTA_DST:
+            route->plen = GET_PLEN(rtm->rtm_dst_len);
             COPY_ADDR(route->prefix, RTA_DATA(rta));
             break;
+        case RTA_SRC:
+            route->src_plen = GET_PLEN(rtm->rtm_src_len);
+            COPY_ADDR(route->src_prefix, RTA_DATA(rta));
+            break;
         case RTA_GATEWAY:
             COPY_ADDR(route->gw, RTA_DATA(rta));
             break;
@@ -1106,6 +1145,7 @@ parse_kernel_route_rta(struct rtmsg *rtm, int len, struct kernel_route *route)
         rta = RTA_NEXT(rta, len);
     }
 #undef COPY_ADDR
+#undef GET_PLEN
 
     int i;
     for(i = 0; i < import_table_count; i++)
@@ -1120,6 +1160,7 @@ print_kernel_route(int add, int protocol, int type,
 {
     char ifname[IFNAMSIZ];
     char addr_prefix[INET6_ADDRSTRLEN];
+    char src_addr_prefix[INET6_ADDRSTRLEN];
     char addr_gw[INET6_ADDRSTRLEN];
 
     if(!inet_ntop(AF_INET6, route->prefix,
@@ -1130,6 +1171,21 @@ print_kernel_route(int add, int protocol, int type,
         return;
     }
 
+    if(route->src_plen >= 0) {
+        if(!inet_ntop(AF_INET6, route->src_prefix,
+                      src_addr_prefix, sizeof(src_addr_prefix))) {
+            kdebugf("Couldn't format kernel route for printing.");
+            return;
+        }
+
+        kdebugf("%s kernel route: dest: %s/%d gw: %s metric: %d if: %s "
+                "(proto: %d, type: %d, from: %s/%d)",
+                add == RTM_NEWROUTE ? "Add" : "Delete",
+                addr_prefix, route->plen, addr_gw, route->metric, ifname,
+                protocol, type, src_addr_prefix, route->src_plen);
+        return;
+    }
+
     kdebugf("%s kernel route: dest: %s/%d gw: %s metric: %d if: %s "
            "(proto: %d, type: %d)",
            add == RTM_NEWROUTE ? "Add" : "Delete",
@@ -1174,9 +1230,6 @@ filter_kernel_routes(struct nlmsghdr *nh, void *data)
     if(rtm->rtm_protocol == RTPROT_BABEL)
         return 0;
 
-    if(rtm->rtm_src_len != 0)
-        return 0;
-
     /* Ignore cached routes, advertised by some kernels (linux 3.x). */
     if(rtm->rtm_flags & RTM_F_CLONED)
         return 0;
@@ -1190,7 +1243,8 @@ filter_kernel_routes(struct nlmsghdr *nh, void *data)
     if(rc < 0)
         return 0;
 
-    if(martian_prefix(current_route->prefix, current_route->plen))
+    if(martian_prefix(current_route->prefix, current_route->plen) ||
+       martian_prefix(current_route->src_prefix, current_route->src_plen))
         return 0;
 
     /* Ignore default unreachable routes; no idea where they come from. */
@@ -1219,6 +1273,7 @@ kernel_routes(struct kernel_route *routes, int maxroutes)
     int found = 0;
     void *data[3] = { &maxr, routes, &found };
     int families[2] = { AF_INET6, AF_INET };
+    char rule_exists[SRC_TABLE_NUM] = {0};
     struct rtgenmsg g;
 
     if(!nl_setup) {
@@ -1244,9 +1299,19 @@ kernel_routes(struct kernel_route *routes, int maxroutes)
 
         rc = netlink_read(&nl_command, NULL, 1,
                           filter_kernel_routes, (void *)data);
-
         if(rc < 0)
             return -1;
+
+        rc = netlink_send_dump(RTM_GETRULE, &g, sizeof(g));
+        if(rc < 0)
+            return -1;
+
+        rc = netlink_read(&nl_command, NULL, 1,
+                          filter_kernel_rules, rule_exists);
+        if(rc < 0)
+            return -1;
+
+        install_missing_rules(rule_exists, families[i] == AF_INET);
     }
 
     return found;
@@ -1341,6 +1406,10 @@ filter_link(struct nlmsghdr *nh, void *data)
     return 0;
 }
 
+/* If data is null, takes all addresses.  If data is not null, takes
+   either link-local or global addresses depending of the value of
+   data[4]. */
+
 static int
 filter_addresses(struct nlmsghdr *nh, void *data)
 {
@@ -1380,7 +1449,7 @@ filter_addresses(struct nlmsghdr *nh, void *data)
     if(rc < 0)
         return 0;
 
-    if(ll == !IN6_IS_ADDR_LINKLOCAL(&addr))
+    if(data && ll == !IN6_IS_ADDR_LINKLOCAL(&addr))
         return 0;
 
     if(ifindex && ifa->ifa_index != ifindex)
@@ -1429,6 +1498,12 @@ filter_netlink(struct nlmsghdr *nh, void *data)
         if(changed && rc > 0)
             *changed |= CHANGE_ADDR;
         return rc;
+    case RTM_NEWRULE:
+    case RTM_DELRULE:
+        rc = filter_kernel_rules(nh, NULL);
+        if(changed && rc > 0)
+            *changed |= CHANGE_RULE;
+        return rc;
     default:
         kdebugf("filter_netlink: unexpected message type %d\n",
                 nh->nlmsg_type);
@@ -1504,3 +1579,430 @@ kernel_callback(int (*fn)(int, void*), void *closure)
 
     return 0;
 }
+
+
+/* Routing table's rules */
+
+static int
+add_rule(int prio, const unsigned char *src_prefix, int src_plen, int table)
+{
+    char buffer[64] = {0}; /* 56 needed */
+    struct nlmsghdr *message_header = (void*)buffer;
+    struct rtmsg *message = NULL;
+    struct rtattr *current_attribute = NULL;
+    int is_v4 = v4mapped(src_prefix);
+    int addr_size = is_v4 ? sizeof(struct in_addr) : sizeof(struct in6_addr);
+
+    kdebugf("Add rule v%c prio %d from %s\n", is_v4 ? '4' : '6', prio,
+            format_prefix(src_prefix, src_plen));
+
+    if(is_v4) {
+        src_prefix += 12;
+        src_plen -= 96;
+        if(src_plen < 0) {
+            errno = EINVAL;
+            return -1;
+        }
+    }
+
+#if RTA_ALIGNTO != NLMSG_ALIGNTO
+#error "RTA_ALIGNTO != NLMSG_ALIGNTO"
+#endif
+
+    /* Set the header */
+    message_header->nlmsg_flags = NLM_F_REQUEST | NLM_F_CREATE | NLM_F_EXCL;
+    message_header->nlmsg_type  = RTM_NEWRULE;
+    message_header->nlmsg_len   = NLMSG_ALIGN(sizeof(struct nlmsghdr));
+
+    /* Append the message */
+    message = NLMSG_DATA(message_header);
+    message->rtm_family = is_v4 ? AF_INET : AF_INET6;
+    message->rtm_dst_len = 0;
+    message->rtm_src_len = src_plen;
+    message->rtm_tos = 0;
+    message->rtm_table = table;
+    message->rtm_protocol = RTPROT_BABEL;
+    message->rtm_scope = RT_SCOPE_UNIVERSE;
+    message->rtm_type = RTN_UNICAST;
+    message->rtm_flags = 0;
+    message_header->nlmsg_len += NLMSG_ALIGN(sizeof(struct rtmsg));
+
+    /* Append each attribute */
+    current_attribute = RTM_RTA(message);
+    /* prio */
+    current_attribute->rta_len = RTA_LENGTH(sizeof(int));
+    current_attribute->rta_type = FRA_PRIORITY;
+    *(int*)RTA_DATA(current_attribute) = prio;
+
+    message_header->nlmsg_len += current_attribute->rta_len;
+    current_attribute = (void*)
+        ((char*)current_attribute) + current_attribute->rta_len;
+
+    /* src */
+    current_attribute->rta_len = RTA_LENGTH(addr_size);
+    current_attribute->rta_type = FRA_SRC;
+    memcpy(RTA_DATA(current_attribute), src_prefix, addr_size);
+
+    message_header->nlmsg_len += current_attribute->rta_len;
+    current_attribute = (void*)
+        ((char*)current_attribute) + current_attribute->rta_len;
+
+    /* send message */
+    if(message_header->nlmsg_len > 64) {
+        errno = EINVAL;
+        return -1;
+    }
+    return netlink_talk(message_header);
+}
+
+static int
+flush_rule(int prio, int family)
+{
+    char buffer[64] = {0}; /* 36 needed */
+    struct nlmsghdr *message_header = (void*)buffer;
+    struct rtmsg *message = NULL;
+    struct rtattr *current_attribute = NULL;
+
+    memset(buffer, 0, sizeof(buffer));
+
+    kdebugf("Flush rule v%c prio %d\n", family == AF_INET ? '4' : '6', prio);
+
+#if RTA_ALIGNTO != NLMSG_ALIGNTO
+#error "RTA_ALIGNTO != NLMSG_ALIGNTO"
+#endif
+
+    /* Set the header */
+    message_header->nlmsg_flags = NLM_F_REQUEST;
+    message_header->nlmsg_type  = RTM_DELRULE;
+    message_header->nlmsg_len   = NLMSG_ALIGN(sizeof(struct nlmsghdr));
+
+    /* Append the message */
+    message = NLMSG_DATA(message_header);
+    message->rtm_family = family;
+    message->rtm_dst_len = 0;
+    message->rtm_src_len = 0;
+    message->rtm_tos = 0;
+    message->rtm_table = 0;
+    message->rtm_protocol = RTPROT_BABEL;
+    message->rtm_scope = RT_SCOPE_UNIVERSE;
+    message->rtm_type = RTN_UNSPEC;
+    message->rtm_flags = 0;
+    message_header->nlmsg_len += NLMSG_ALIGN(sizeof(struct rtmsg));
+
+    /* Append each attribute */
+    current_attribute = RTM_RTA(message);
+    /* prio */
+    current_attribute->rta_len = RTA_LENGTH(sizeof(int));
+    current_attribute->rta_type = FRA_PRIORITY;
+    *(int*)RTA_DATA(current_attribute) = prio;
+
+    message_header->nlmsg_len += current_attribute->rta_len;
+    current_attribute = (void*)
+        ((char*)current_attribute) + current_attribute->rta_len;
+
+    /* send message */
+    if(message_header->nlmsg_len > 64) {
+        errno = EINVAL;
+        return -1;
+    }
+    return netlink_talk(message_header);
+}
+
+
+/* Source specific functions and data structures */
+
+/* The table used for non-specific routes is "export_table", therefore, we can
+   take the convention of plen == 0 <=> empty table. */
+struct kernel_table {
+    unsigned char src[16];
+    unsigned char plen;
+    unsigned char table;
+};
+
+/* kernel_tables contains informations about the rules we installed. It is an
+   array indexed by: <table priority> - src_table_prio.
+   (First entries are the most specific, since they have priority.) */
+static struct kernel_table kernel_tables[SRC_TABLE_NUM];
+/* used tables is indexed by: <table number> - src_table_idx
+   used_tables[i] == 1 <=> the table number (i + src_table_idx) is used */
+static char used_tables[SRC_TABLE_NUM] = {0};
+static unsigned char src_table_used = 0; /* number of tables used */
+
+static inline int
+change_table_priority(const unsigned char *src, int plen, int table,
+                      int old_prio, int new_prio)
+{
+    int rc;
+    kdebugf("/Swap: ");
+    rc = add_rule(new_prio, src, plen, table);
+    if(rc < 0)
+        return rc;
+    kdebugf("\\Swap: ");
+    return flush_rule(old_prio, v4mapped(src) ? AF_INET : AF_INET6);
+}
+
+/* Return a new table at index [idx] of kernel_tables.  If cell at that index is
+   not free, we need to shift cells (and rules).  If it's full, return NULL. */
+static struct kernel_table *
+insert_table(const unsigned char *src, unsigned short src_plen, int idx)
+{
+    int table;
+    int rc;
+    int i;
+
+    if(idx < 0 || idx >= SRC_TABLE_NUM) {
+        fprintf(stderr, "Incorrect table number %d\n", idx);
+        return NULL;
+    }
+
+    if(src_table_used >= SRC_TABLE_NUM) {
+        kdebugf("All allowed routing tables are used!\n");
+        return NULL;
+    }
+
+    /* find a free table number */
+    for(table = 0; table < SRC_TABLE_NUM; table++)
+        if(!used_tables[table])
+            break;
+    table += src_table_idx;
+
+    /* Create the table's rule at the right place. Shift rules if necessary. */
+    if(kernel_tables[idx].plen != 0) {
+        /* shift right */
+        i = src_table_used;
+        while(i > idx) {
+            i--;
+            rc = change_table_priority(kernel_tables[i].src,
+                                       kernel_tables[i].plen,
+                                       kernel_tables[i].table,
+                                       i + src_table_prio,
+                                       i + 1 + src_table_prio);
+            if(rc < 0) {
+                perror("change_table_priority");
+                return NULL;
+            }
+            kernel_tables[i+1] = kernel_tables[i];
+            kernel_tables[i].plen = 0;
+        }
+    }
+
+    rc = add_rule(idx + src_table_prio, src, src_plen, table);
+    if(rc < 0) {
+        perror("add rule");
+        return NULL;
+    }
+    used_tables[table - src_table_idx] = 1;
+    memcpy(kernel_tables[idx].src, src, 16);
+    kernel_tables[idx].plen = src_plen;
+    kernel_tables[idx].table = table;
+
+    src_table_used++;
+    return &kernel_tables[idx];
+}
+
+/* Sorting kernel_tables in a well ordered fashion will increase code complexity
+   and decrease performances, because more rule shifs will be required, so more
+   system calls invoked. */
+static int
+find_table_slot(const unsigned char *src, unsigned short src_plen,
+                 int *new_return)
+ {
+    struct kernel_table *kt = NULL;
+    int i;
+    *new_return = -1;
+
+    for(i = 0; i < SRC_TABLE_NUM; i++) {
+        kt = &kernel_tables[i];
+        if(kt->plen == 0)
+            goto new_table_here; /* empty table here */
+        switch(prefix_cmp(src, src_plen, kt->src, kt->plen)) {
+        case PST_LESS_SPECIFIC:
+        case PST_DISJOINT:
+            continue; /* try to find a comparable route entry. */
+        case PST_MORE_SPECIFIC:
+            goto new_table_here;
+        case PST_EQUALS:
+            return i;
+        }
+    }
+
+    return -1;
+ new_table_here:
+    *new_return = i;
+    return -1;
+ }
+
+static int
+find_table(const unsigned char *src, unsigned short src_plen)
+{
+    struct kernel_table *kt = NULL;
+    int i, new_i;
+
+    if(src_plen == 0 ||
+       (has_ipv6_subtrees && (src_plen < 96 || !v4mapped(src)))) {
+        fprintf(stderr, "Find_table called for route handled by kernel "
+                "(this shouldn't happen).");
+        return -1;
+    }
+
+    i = find_table_slot(src, src_plen, &new_i);
+    if(i < 0) {
+        if(new_i < 0)
+            return -1;
+        kt = insert_table(src, src_plen, new_i);
+    } else {
+        kt = &kernel_tables[i];
+    }
+    return kt == NULL ? -1 : kt->table;
+}
+
+static void
+release_tables(void)
+{
+    int i;
+    for(i = 0; i < SRC_TABLE_NUM; i++) {
+        if(kernel_tables[i].plen != 0) {
+            flush_rule(i + src_table_prio,
+                       v4mapped(kernel_tables[i].src) ? AF_INET : AF_INET6);
+            kernel_tables[i].plen = 0;
+        }
+        used_tables[i] = 0;
+    }
+    src_table_used = 0;
+}
+
+static int
+filter_kernel_rules(struct nlmsghdr *nh, void *data)
+{
+    int i, len, has_priority = 0;
+    unsigned int rta_len;
+    struct rtmsg *rtm = NULL;
+    struct rtattr *rta = NULL;
+    int is_v4 = 0;
+    char *rule_exists = data;
+    unsigned char src[16];
+    unsigned char src_plen;
+    unsigned int table, priority = 0xFFFFFFFF;
+
+    len = nh->nlmsg_len;
+
+    rtm = (struct rtmsg*)NLMSG_DATA(nh);
+    len -= NLMSG_LENGTH(0);
+
+    src_plen = rtm->rtm_src_len;
+    table = rtm->rtm_table;
+
+    if(rtm->rtm_family != AF_INET && rtm->rtm_family != AF_INET6) {
+        kdebugf("filter_rules: Unknown family: %d\n", rtm->rtm_family);
+        return -1;
+    }
+    is_v4 = rtm->rtm_family == AF_INET;
+
+    rta = RTM_RTA(rtm);
+    len -= NLMSG_ALIGN(sizeof(*rtm));
+
+#define GET_PLEN(p) (rtm->rtm_family == AF_INET) ? p + 96 : p
+#define COPY_ADDR(d, s)                                                 \
+    do {                                                                \
+        if(!is_v4) {                                                    \
+            if(UNLIKELY(rta_len < 16)) {                                \
+                fprintf(stderr, "filter_rules: truncated message.");    \
+                return -1;                                              \
+            }                                                           \
+            memcpy(d, s, 16);                                           \
+        }else {                                                         \
+            if(UNLIKELY(rta_len < 4)) {                                 \
+                fprintf(stderr, "filter_rules: truncated message.");    \
+                return -1;                                              \
+            }                                                           \
+            v4tov6(d, s);                                               \
+        }                                                               \
+    } while(0)
+
+    for(; RTA_OK(rta, len); rta = RTA_NEXT(rta, len)) {
+        rta_len = RTA_PAYLOAD(rta);
+        switch(rta->rta_type) {
+        case FRA_UNSPEC: break;
+        case FRA_SRC:
+            src_plen = GET_PLEN(rtm->rtm_src_len);
+            COPY_ADDR(src, RTA_DATA(rta));
+            break;
+        case FRA_PRIORITY:
+            priority = *(unsigned int*)RTA_DATA(rta);
+            has_priority = 1;
+            break;
+        case FRA_TABLE:
+            table = *(int*)RTA_DATA(rta);
+            break;
+        default:
+            kdebugf("filter_rules: Unknown rule attribute: %d.\n",
+                    rta->rta_type);
+            break;
+        }
+    }
+#undef COPY_ADDR
+#undef GET_PLEN
+
+    kdebugf("filter_rules: from %s prio %d table %d\n",
+            format_prefix(src, src_plen), priority, table);
+
+    if(martian_prefix(src, src_plen) || !has_priority)
+        return 0;
+
+    i = priority - src_table_prio;
+
+    if(i < 0 || SRC_TABLE_NUM <= i)
+        return 0;
+
+    /* There is an unexpected change on one of our rules. */
+    if(!rule_exists)
+        return 1;
+
+    if(prefix_cmp(src, src_plen,
+                  kernel_tables[i].src, kernel_tables[i].plen)
+       == PST_EQUALS &&
+       table == kernel_tables[i].table &&
+       !rule_exists[i]) {
+        rule_exists[i] = 1;
+    } else {
+        int rc;
+        do {
+            rc = flush_rule(i + src_table_prio, is_v4 ? AF_INET : AF_INET6);
+        } while(rc >= 0);
+        /* flush unexpected rules, but keep information in kernel_tables[i].  It
+            will be used afterwards to reinstall the rule. */
+        if(errno != ENOENT && errno != EEXIST)
+            fprintf(stderr,
+                    "filter_rules: cannot remove from %s prio %d table %d"
+                    "(%s)\n",
+                    format_prefix(src, src_plen), priority, table,
+                    strerror(errno));
+        rule_exists[i] = 0;
+    }
+
+    return 1;
+}
+
+/* This functions should be executed wrt the code just bellow: [rule_exists]
+   contains is a boolean array telling whether the rules we should have
+   installed in the kernel are installed or not.  If they aren't, then reinstall
+   them (this can append when rules are modified by third parties). */
+
+static void
+install_missing_rules(char rule_exists[SRC_TABLE_NUM], int v4)
+{
+    int i, rc;
+    for(i = 0; i < SRC_TABLE_NUM; i++)
+        if(v4mapped(kernel_tables[i].src) == v4 &&
+           !rule_exists[i] && kernel_tables[i].plen != 0) {
+            rc = add_rule(i + src_table_prio, kernel_tables[i].src,
+                          kernel_tables[i].plen, kernel_tables[i].table);
+            if(rc < 0)
+                fprintf(stderr,
+                        "install_missing_rules: "
+                        "Cannot install rule: table %d prio %d from %s\n",
+                        kernel_tables[i].table, i + src_table_prio,
+                        format_prefix(kernel_tables[i].src,
+                                      kernel_tables[i].plen));
+        }
+}
